@@ -15,10 +15,14 @@ use overload '""' => \&_stringify,
              $] < 5.037 ? ('~~' => \&_matches) : ();  # fully deprecated, so cannot be overloaded
 use match::simple ();
 
-our $VERSION = "1.12";
+our $VERSION = "1.13";
+
+our $MESSAGE;            # global var for last message from _matches()
+our $MAX_DEEP = 100;     # limit for recursive calls to inspect()
+our $GLOBAL_MSGS;        # table of default messages -- see below method messages()
+our $USE_OLD_MSG_API;    # flag for backward compatibility
+
 
-our $MESSAGE;        # global var for last message from _matches()
-our $MAX_DEEP = 100; # limit for recursive calls to inspect()
 
 #----------------------------------------------------------------------
 # exports
@@ -49,7 +53,7 @@ BEGIN {
 
 # setup exports through Sub::Exporter API
 use Sub::Exporter -setup => {
-  exports    => [ 'node_from_path',                                          # no longer documented, but still present for backwards compat
+  exports    => [ 'node_from_path',                                          # no longer used, but still present for backwards compat
                   (map {$_ => \&_wrap_domain          } @CONSTRUCTORS  ),
                   (map {$_ => \&_wrap_shortcut_options} keys %SHORTCUTS) ],
   groups     => { constructors => \@CONSTRUCTORS,
@@ -103,6 +107,13 @@ sub _wrap_shortcut_options {
 # messages
 #----------------------------------------------------------------------
 
+sub _msg_bool { # small closure generator for various messages below
+  my ($must_be, $if_true, $if_false) = @_;
+  return sub {my ($name, $msg_id, $expected) = @_;
+              "$name: $must_be " . ($expected ? $if_true : $if_false)};
+}
+
+
 my $builtin_msgs = {
   english => {
     Generic => {
@@ -111,20 +122,20 @@ my $builtin_msgs = {
       TOO_SMALL     => "smaller than minimum '%s'",
       TOO_BIG       => "bigger than maximum '%s'",
       EXCLUSION_SET => "belongs to exclusion set",
-      MATCH_TRUE    => "data true/false",
+      MATCH_TRUE    => _msg_bool("must be", "true", "false"),
       MATCH_ISA     => "is not a '%s'",
       MATCH_CAN     => "does not have method '%s'",
       MATCH_DOES    => "does not do '%s'",
-      MATCH_BLESSED => "data blessed/unblessed",
-      MATCH_PACKAGE => "data is/is not a package",
-      MATCH_REF     => "is/is not a reference",
+      MATCH_BLESSED => _msg_bool("must be", "blessed", "unblessed"),
+      MATCH_PACKAGE => _msg_bool("must be", "a package", "a non-package"),
+      MATCH_REF     => _msg_bool("must be", "a reference", "a non-reference"),
       MATCH_SMART   => "does not smart-match '%s'",
-      MATCH_ISWEAK  => "weak/strong reference",
-      MATCH_READONLY=> "readonly data",
-      MATCH_TAINTED => "tainted/untainted",
+      MATCH_ISWEAK  => _msg_bool("must be", "a weak reference", "a strong reference"),
+      MATCH_READONLY=> _msg_bool("must be", "readonly", "non-readonly"),
+      MATCH_TAINTED => _msg_bool("must be", "tainted", "untainted"),
     },
     Whatever => {
-      MATCH_DEFINED => "data defined/undefined",
+      MATCH_DEFINED => _msg_bool("must be", "defined", "undefined"),
     },
     Num    => {INVALID => "invalid number",},
     Date   => {INVALID => "invalid date",},
@@ -155,20 +166,20 @@ my $builtin_msgs = {
       TOO_SMALL     => "plus petit que le minimum '%s'",
       TOO_BIG       => "plus grand que le maximum '%s'",
       EXCLUSION_SET => "fait partie des valeurs interdites",
-      MATCH_TRUE    => "donne vraie/fausse",
+      MATCH_TRUE    => _msg_bool("doit tre", "vrai", "faux"),
       MATCH_ISA     => "n'est pas un  '%s'",
       MATCH_CAN     => "n'a pas la mthode '%s'",
       MATCH_DOES    => "ne se comporte pas comme un '%s'",
-      MATCH_BLESSED => "donne blessed/unblessed",
-      MATCH_PACKAGE => "est/n'est pas un package",
-      MATCH_REF     => "est/n'est pas une rfrence",
+      MATCH_BLESSED => _msg_bool("doit tre", "blessed", "unblessed"),
+      MATCH_PACKAGE => _msg_bool("doit tre", "un package", "un non-package"),
+      MATCH_REF     => _msg_bool("doit tre", "une rfrence", "une non-rfrence"),
       MATCH_SMART   => "n'obit pas au smart-match '%s'",
-      MATCH_ISWEAK  => "rfrence weak/strong",
-      MATCH_READONLY=> "donne readonly",
-      MATCH_TAINTED => "tainted/untainted",
+      MATCH_ISWEAK  => _msg_bool("doit tre", "une weak reference", "une strong reference"),
+      MATCH_READONLY=> _msg_bool("doit tre", "readonly", "non-readonly"),
+      MATCH_TAINTED => _msg_bool("doit tre", "tainted", "untainted"),
     },
     Whatever => {
-      MATCH_DEFINED => "donne dfinie/non dfinie",
+      MATCH_DEFINED => _msg_bool("doit tre", "dfini", "non-dfini"),
     },
     Num    => {INVALID => "nombre incorrect",},
     Date   => {INVALID => "date incorrecte",},
@@ -200,7 +211,7 @@ foreach my $language (keys %$builtin_msgs) {
 }
 
 # default messages : english
-my $global_msgs = $builtin_msgs->{english};
+$GLOBAL_MSGS = $builtin_msgs->{english};
 
 #----------------------------------------------------------------------
 # PUBLIC METHODS
@@ -211,7 +222,7 @@ sub messages { # private class method
   croak "messages() is a class method in Data::Domain" 
     if ref $class or $class ne 'Data::Domain';
 
-  $global_msgs = (ref $new_messages) ? $new_messages 
+  $GLOBAL_MSGS = (ref $new_messages) ? $new_messages 
                                      : $builtin_msgs->{$new_messages}
     or croak "no such builtin messages ($new_messages)";
 }
@@ -349,38 +360,48 @@ sub _check_returns {
 # METHODS FOR INTERNAL USE
 #----------------------------------------------------------------------
 
-
 sub msg {
   my ($self, $msg_id, @args) = @_;
   my $msgs     = $self->{-messages};
   my $subclass = $self->subclass;
   my $name     = $self->{-name} || $subclass;
-  my $msg;
+
+  # if using a coderef, these args will be passed to it
+  my @msgs_call_args = ($name, $msg_id, @args);
+  shift @msgs_call_args if $USE_OLD_MSG_API;
 
   # perl v5.22 and above warns if there are too many @args for sprintf.
   # The line below prevents that warning
   no if $] ge '5.022000', warnings => 'redundant';
 
-  # if there is a user_defined message, return it
+  # if there is a user-defined message, return it
   if (defined $msgs) { 
     for (ref $msgs) {
-      /^CODE/ and return $msgs->($msg_id, @args); # user function
-      /^$/    and return "$name: $msgs";          # user constant string
-      /^HASH/ and do { $msg =  $msgs->{$msg_id}   # user hash of msgs
-                         and return sprintf "$name: $msg", @args;
-                       last; # not found in this hash - revert to $global_msgs
+      /^CODE/ and return $msgs->(@msgs_call_args);                # user function
+      /^$/    and return "$name: $msgs";                          # user constant string
+      /^HASH/ and do { if (my $msg_string =  $msgs->{$msg_id}) {  # user hash of msgs
+                         return sprintf "$name: $msg_string", @args;
+                       }
+                       else {
+                         last; # not found in this hash - revert to $GLOBAL_MSGS below
+                       }
                      };
-      croak "invalid -messages option";           # otherwise
+      # otherwise
+      croak "-messages option should be a coderef, a hashref or a sprintf string";
     }
   }
 
-  # otherwise, try global messages
-  return $global_msgs->($msg_id, @args)      if ref $global_msgs eq 'CODE';
-  $msg = $global_msgs->{$subclass}{$msg_id}  # otherwise
-      || $global_msgs->{Generic}{$msg_id}
+  # there was no user-defined message, so use global messages
+  if (ref $GLOBAL_MSGS eq 'CODE') {
+    return $GLOBAL_MSGS->(@msgs_call_args);
+  }
+  else {
+    my $msg_entry = $GLOBAL_MSGS->{$subclass}{$msg_id}
+                  || $GLOBAL_MSGS->{Generic}{$msg_id}
      or croak "no error string for message $msg_id";
-
-  return sprintf "$name: $msg", @args;
+    return ref $msg_entry eq 'CODE' ? $msg_entry->(@msgs_call_args)
+                                    : sprintf "$name: $msg_entry", @args;
+  }
 }
 
 
@@ -442,12 +463,11 @@ sub _build_subdomain {
   elsif (does($domain, 'CODE')) {
     # this is a lazy domain, need to call the coderef to get a real domain
     $domain = try   {$domain->($context)} 
-      catch {   # remove "at source_file, line ..." from error message
-        (my $error_msg = $_) =~ s/\bat\b.*//s;
-        # return an empty domain that reports the error message
-        Data::Domain::Empty->new(-name => "domain parameters",
-                                 -messages => $error_msg);
-      };
+              catch {(my $error_msg = $_) =~ s/\bat\b.*//s; # remove "at source_file, line ..." from error message
+                     # return an empty domain that reports the error message
+                     Data::Domain::Empty->new(-name     => "domain parameters",
+                                              -messages => $error_msg);
+                   };
     # did we really get a domain ?
     does($domain, "Data::Domain")
       or croak "lazy domain coderef returned an invalid domain";
@@ -467,6 +487,15 @@ sub _build_subdomain {
 }
 
 
+sub _is_proper_subdomain {
+  my ($self, $domain) = @_;
+  return does($_, 'Data::Domain') || does($_, 'CODE') || !ref $_;
+}
+
+
+
+
+
 #----------------------------------------------------------------------
 # UTILITY FUNCTIONS (NOT METHODS) 
 #----------------------------------------------------------------------
@@ -505,7 +534,7 @@ sub _parse_args {
 }
 
 
-sub node_from_path { # no longer documented, but still present for backwards compat
+sub node_from_path { # no longer used (replaced by Data::Reach); but still present for backwards compat
   my ($root, $path0, @path) = @_;
   return $root if not defined $path0;
   return undef if not defined $root;
@@ -1053,6 +1082,10 @@ sub new {
       croak "$bound does not match -items"
       if $self->{$bound} and $self->{$bound} < @{$self->{-items}};
     }
+
+    # check that all items are associated to proper subdomains
+    my @invalid_fields = grep {!$self->_is_proper_subdomain($self->{-items}[$_])} 0 .. $#{$self->{-items}};
+    croak "invalid subdomain for field: ", join ", ", @invalid_fields  if @invalid_fields;
   }
 
   # check that -all or -any are domains or lists of domains
@@ -1161,10 +1194,11 @@ our @ISA = 'Data::Domain';
 
 sub new {
   my $class = shift;
-  my @options = qw/-fields -exclude -keys -values/;
+  my @options = qw/-fields -exclude -keys -values -may_ignore/;
   my $self = Data::Domain::_parse_args(\@_, \@options, -fields => 'arrayref');
   bless $self, $class;
 
+  # parse the -fields option
   my $fields = $self->{-fields} || [];
   if (does($fields, 'ARRAY')) {
     # transform arrayref into hashref plus an ordered list of keys
@@ -1178,18 +1212,22 @@ sub new {
   }
   elsif (does($fields, 'HASH')) {
     # keep given hashref, add list of keys
-    $self->{-fields_list} = [keys %$fields];
+    $self->{-fields_list} = [sort keys %$fields];
   }
   else {
     croak "invalid data for -fields option";
   }
 
-  # check that -exclude is an arrayref or a regex or a string
-  if (my $exclude = $self->{-exclude}) {
-    does($exclude, 'ARRAY') || does($exclude, 'Regexp') || !ref($exclude)
-      or croak "invalid data for -exclude option";
-  }
+  # check that all fields are associated to proper subdomains
+  my @invalid_fields = grep {!$self->_is_proper_subdomain($self->{-fields}{$_})} @{$self->{-fields_list}};
+  croak "invalid subdomain for field: ", join ", ", @invalid_fields  if @invalid_fields;
 
+  # check that -exclude and -may_ignore are an arrayref or a regex or a string
+  for my $opt (qw/-exclude -may_ignore/) {
+    my $val = $self->{$opt} or next;
+    does($val, 'ARRAY') || does($val, 'Regexp') || !ref($val)
+      or croak "invalid data for $opt option";
+  }
 
   # check that -keys or -values are List domains
   for my $arg (qw/-keys -values/) {
@@ -1203,6 +1241,10 @@ sub new {
 }
 
 
+
+
+
+
 sub _inspect {
   my ($self, $data, $context) = @_;
   no warnings 'recursion';
@@ -1216,10 +1258,8 @@ sub _inspect {
   # check if there are any forbidden fields
   if (my $exclude = $self->{-exclude}) {
     my @other_fields = grep {!$self->{-fields}{$_}} keys %$data;
-    my @wrong_fields = match::simple::match($exclude, ['*', 'all'])
-                       ? @other_fields
-                       : grep {match::simple::match($_, $exclude)} @other_fields;
-    $msgs{-exclude} = $self->msg(FORBIDDEN_FIELD => join ", ", map {"'$_'"} sort @wrong_fields)
+    my @wrong_fields = grep {$self->_field_matches(-exclude => $_)} @other_fields;
+    $msgs{-exclude}  = $self->msg(FORBIDDEN_FIELD => join ", ", map {"'$_'"} sort @wrong_fields)
       if @wrong_fields;
   }
 
@@ -1231,7 +1271,9 @@ sub _inspect {
   local $context->{flat} = {%{$context->{flat}}, %$data};
 
   # check fields of the domain
+ FIELD:
   foreach my $field (@{$self->{-fields_list}}) {
+    next FIELD if not exists $data->{$field} and $self->_field_matches(-may_ignore => $field);
     local $context->{path} = [@{$context->{path}}, $field];
     my $field_spec = $self->{-fields}{$field};
     my $subdomain  = $self->_build_subdomain($field_spec, $context);
@@ -1258,6 +1300,16 @@ sub _inspect {
   return keys %msgs ? \%msgs : undef;
 }
 
+sub _field_matches {
+  my ($self, $spec, $field) = @_;
+
+  my $spec_content = $self->{$spec};
+  return $spec_content && (match::simple::match($spec_content, ['*', 'all'])
+                           ||
+                           match::simple::match($field, $spec_content));
+}
+
+
 #======================================================================
 package Data::Domain::One_of;
 #======================================================================
@@ -1369,15 +1421,16 @@ Data::Domain - Data description and validation
                  });
 
   # examples of subroutines for specialized domains
-  sub Phone   { String(-regex    => qr/^\+?[0-9() ]+$/, 
-                       -messages => "Invalid phone number", @_) }
-  sub Email   { String(-regex    => qr/^[-.\w]+\@[\w.]+$/,
-                       -messages => "Invalid email", @_) }
-  sub Contact { Struct(-fields => [name   => String,
-                                   phone  => Phone,
-                                   mobile => Phone(-name => 'Mobile',
-                                                   -optional => 1),
-                                   emails => List(-all => Email)   ], @_) }
+  sub Phone         { String(-regex    => qr/^\+?[0-9() ]+$/, 
+                             -messages => "Invalid phone number", @_) }
+  sub Email         { String(-regex    => qr/^[-.\w]+\@[\w.]+$/,
+                             -messages => "Invalid email", @_) }
+  sub Contact       { Struct(-fields => [name   => String,
+                                         phone  => Phone,
+                                         mobile => Phone(-name => 'Mobile',
+                                                         -optional => 1),
+                                         emails => List(-all => Email)   ], @_) }
+  sub UpdateContact { Contact(-may_ignore => '*', @_) }
 
   # lazy subdomain
   $domain = Struct(
@@ -1401,7 +1454,7 @@ Data::Domain - Data description and validation
 
 =head1 DESCRIPTION
 
-A data domain is a description of a set of values, either scalar or
+A I<data domain> is a description of a set of values, either scalar or
 structured (arrays or hashes).  The description can include many
 constraints, like minimal or maximal values, regular expressions,
 required fields, forbidden fields, and also contextual
@@ -1430,6 +1483,16 @@ the companion module L<Test::InDomain>.
 There are several other packages in CPAN doing data validation; these
 are briefly listed in the L</"SEE ALSO"> section.
 
+=head1 COMPATIBILITY WARNING : API CHANGE FOR CODEREFS
+
+Starting with version 1.13, the API for calling message coderefs has
+changed and is now in the form
+
+  $coderef->($domain_name, $msg_id, @args);
+
+See section L<Backward compatibility for coderefs> for a workaround.
+
+
 =head1 EXPORTS
 
 =head2 Domain constructors
@@ -1538,7 +1601,7 @@ So in short, the "default option" is syntactic sugar for using positional
 parameters instead of named parameters.
 
 Each domain constructor has its own list of available options; these will be
-presented below, together with each subclass (for example options for
+presented with each subclass (for example options for
 setting minimal/maximal values, regular expressions, string length,
 etc.).  However, there are also some generic options, available in
 every domain constructor; these are listed here, in several categories.
@@ -2134,6 +2197,9 @@ a string starting with C<foo> I<or> a number between 1 and 10.
   my $domain = Struct(-fields  => [foo => Int, bar => String],
                       -exclude => '*'); # only 'foo' and 'bar', nothing else
   
+  my $domain = Struct(-fields     => [foo => Int, bar => String],
+                      -may_ignore => '*'); # will not complain for missing fields
+  
   my $domain = Struct(-keys   => List(-all => String(qr/^[abc])),
                       -values => List(-all => Int));
 
@@ -2144,7 +2210,7 @@ Options are:
 
 =item -fields
 
-Supplies a list of keys with their associated domains. The list might
+Supplies a list of fields (hash keys) with their associated domains. The list might
 be given either as a hashref or as an arrayref.  Specifying it as an
 arrayref is useful for controlling the order in which field checks
 will be performed; this may make a difference when there are context
@@ -2154,12 +2220,36 @@ L<"LAZY CONSTRUCTORS"|/"LAZY CONSTRUCTORS (CONTEXT DEPENDENCIES)"> below ).
 
 =item -exclude
 
-Specifies which keys are not allowed in the structure. The exclusion
-may be specified as an arrayref of key names, as a compiled regular
+Specifies which fields are not allowed in the structure. The exclusion
+may be specified as an arrayref of field names, as a compiled regular
 expression, or as the string constant 'C<*>' or 'C<all>' (meaning that
-no key will be allowed except those explicitly listed in the
+no hash key will be allowed except those explicitly listed in the
 C<-fields> option.
 
+
+=item -may_ignore
+
+Specifies which fields may be ignored by the domain, i.e. may not exist
+in the inspected structure. Like for C<-exclude>, this option can be specified
+as an arrayref of field names, as a compiled regular
+expression, or as the string constant 'C<*>' or 'C<all>'.
+Absent fields will not generate errors if their name matches this specification.
+This is especially useful when your application needs to distinguish between
+an INSERT operation, where all fields must be present, and an UPDATE operation,
+where only a subset of fields are updated -- see the example in the L</SYNOPSIS>.
+
+Another way is to use the C<-optional> flag in domains associated with fields; but there
+is a subtle difference : C<-optional> accepts both missing keys or keys
+containing C<undef>, while C<-may_ignore> only accepts missing keys. Consider :
+
+  Struct(
+    -fields     => {a => Int, b => Int(-optional => 1), c => Int, d => Str},
+    -may_ignore => [qw/c d/],
+  )
+
+In this domain, C<a> must always be present, C<b> may be absent or may be undef, C<c> and C<d>
+may be absent but if present cannot be undef.
+
 =item -keys
 
 Specifies a List domain, for inspecting the list of keys in the hash.
@@ -2483,17 +2573,7 @@ validation error within the domain
 =item *
 
 a hashref : keys of the hash should be message identifiers, and
-values should be the associated error strings.
-
-=item *
-
-a coderef : the referenced subroutine is called, and should
-return the error string. The called subroutine receives
-the message identifier as argument.
-
-=back
-
-Here is an example :
+values should be the associated error strings. Here is an example :
 
   sub Phone { 
     String(-regex      => qr/^\+?[0-9() ]+$/, 
@@ -2505,6 +2585,15 @@ Here is an example :
   }
 
 
+=item *
+
+a coderef : the referenced subroutine is called, and should
+return the error string. The called subroutine receives
+as arguments: C<< ($domain_name, $message_id, @optional_domain_args) >>
+
+=back
+
+
 =head2 The C<messages> class method
 
 Default strings associated with message identifiers are stored in a
@@ -2521,18 +2610,18 @@ The same method can also receive  a custom table.
   Data::Domain->messages($custom_table);
 
 This should be a two-level hashref : first-level entries in the hash
-correspond to C<Data::Domain> subclasses (i.e C<< Num => {...} >>, C<<
-String => {...} >>), or to the constant C<Generic>; for each of those,
+correspond to C<Data::Domain> subclasses (i.e C<< Num => {...} >>,
+C<< String => {...} >>), or to the constant C<Generic>; for each of those,
 the second-level entries should correspond to message identifiers as
 specified in the doc for each subclass (for example C<TOO_SHORT>,
-C<NOT_A_HASH>, etc.).  Values should be strings suitable to be fed to
-L<sprintf>.  Look at C<$builtin_msgs> in the source code to see an
+C<NOT_A_HASH>, etc.).  Values should be either strings suitable to be fed to
+L<sprintf>, or coderefs. Look at C<$builtin_msgs> in the source code to see an
 example.
 
 Finally, it is also possible to write your own message generation 
 handler : 
 
-  Data::Domain->messages(sub {my ($msg_id, @args) = @_;
+  Data::Domain->messages(sub {my ($domain_name, $msg_id, @args) = @_;
                               return "you just got it wrong ($msg_id)"});
 
 What is received in 
@@ -2540,6 +2629,36 @@ C<@args> depends on which validation rule is involved;
 it can be for example the minimal or maximal bounds,
 or the regular expression being checked.
 
+Clearly this class method has a global side-effect. In most cases
+this is exactly what is expected. However it is possible to limit
+the impact by localizing the C<$msgs> class variable :
+
+  { local $Data::Domain::GLOBAL_MSGS;
+    Data::Domain->messages($custom_table);
+
+    check_my_data(...);
+  }
+  # end of block; Data::Domain is back to the original messages table
+
+
+
+=head2 Backward compatibility for coderefs
+
+In the current version of this module, message coderefs are called as
+
+  $coderef->($domain_name, $msg_id, @args);
+
+Versions prior to 1.13 used a different API where the $domain_name was not available :
+
+  $coderef->($msg_id, @args);
+
+So for clients that were using coderefs in versions prior to 1.13, this is an
+B<incompatible change>. Backward compatibility can be restored by setting a
+global variable to a true value :
+
+  $Data::Domain::USE_OLD_MSG_API = 1;
+
+
 =head2 The C<-name> option to domain constructors
 
 The name of the domain is prepended in front of error 
@@ -2715,7 +2834,20 @@ borrowed from L<Test::Deep> and from L<Moose::Manual::Types>.
 
 =head1 ACKNOWLEDGEMENTS
 
-Thanks to David Cantrell and Gabor Szabo for their help on issues related to smartmatch deprecation.
+Thanks to
+
+=over
+
+=item *
+
+David Cantrell and Gabor Szabo for their help on issues related to smartmatch deprecation.
+
+=item *
+
+David Schmidt (davewood) for suggesting extensions to the Struct() domain.
+
+=back
+
 
 =head1 AUTHOR
 
@@ -2723,7 +2855,7 @@ Laurent Dami, E<lt>dami at cpan.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright 2006, 2007, 2012, 2023 by Laurent Dami.
+Copyright 2006-2023 by Laurent Dami.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
